#!/bin/sh

prog=`basename $0`
localProtocol=file
remoteProtocol=gsiftp
localMachine=localhost

default_port="4568" 
default_serverType="data_server"

echoError ()   { echo "$prog: `date "+%a %D %T"`: ERROR : $@" >&2; }
echoInfo ()    { echo "$prog: `date "+%a %D %T"`: INFO  : $@" >&1; }
echoWarning () { echo "$prog: `date "+%a %D %T"`: WARN  : $@" >&2; }
die () { echoError $@ "... aborting"; exit 1; }
usage () { die "USAGE NOT IMPLEMENTED YET"; }
get_value () { echo $1 | sed "s?^--[a-z|A-Z|0-9|-]*=??g"; }

#Check configuration
[ -n "$GLOBUS_LOCATION" ] || die "GLOBUS_LOCATION not set, setup globus first"

if [ $# -lt 2 ]; then
   echoError "Too few arguments"
   usage
fi

##PARSE ARGUMENTS (sam_cp compliant)
save_args=""
while [ -n "$1" ]; do
   case $1 in
      --server-subject=*)
         serverSubject=`get_value "$1"`
	 echoInfo "Set server subject to: ${serverSubject}"
         ;;

      --port=*)
         port=`get_value "$1"`
         echoInfo "Set port number to: ${port}"
         ;;

      --user-proxy=*)
         export X509_USER_PROXY=`get_value "$1"`
         echoInfo "Set X509_USER_PROXY to ${X509_USER_PROXY}" 
         ;;
                
      --x509-cert-dir=*)
         export X509_CERT_DIR=`get_value "$1"`
         echoInfo "Set X509_CERT_DIR to ${X509_CERT_DIR}" 
         ;;
	
      --server-type=*)
         serverType=`get_value "$1"`
	 echoInfo "Set serverType to ${serverType}"
         ;;
      
      *)
         save_args="$save_args $1"
   esac
   shift
done

extra_args=`echo $save_args |\
   awk 'BEGIN{i=1}{while(i<=NF){if(index($i,"-") == 1){printf("%s ",$i)}; i++; }}'`

if [ -n "$extra_args" ]; then
  echoError "Unrecognized arguments: '$extra_args'"
  usage
fi

# check proxy
user_proxy_subject=`grid-proxy-info -subject 2>&1`
if [ "$?" = "0" ];then
   echoInfo "User Subject: ${user_proxy_subject}"
else
   die $user_proxy_subject
fi

# DETERMINE source and destination files (sam_cp complicant)
first_arg_file=`echo $save_args | awk '{print $1}'`
second_arg_file=`echo $save_args | awk '{print $2}'`
extra_files=`echo $save_args | awk '{print $3}'`
if [ ! -z "$extra_files" ]; then
  echoError "Too many files specified: expecting only 2: '$save_args'"
  usage
fi

colonPresence1=`echo ${first_arg_file} | awk '{print index($1,":")}' -` 	# 0 if no ':'
colonPresence2=`echo ${second_arg_file} | awk '{print index($1,":")}' -` 	# 0 if no ':'

# Check if gridftp needs to perform a PUT or a GET checking whether the
#  colon (:) is in f1 or f2. No third party file transfer allowed
if [ "x$colonPresence1" != "x0" -a "x$colonPresence2" != "x0" ]; then
    echoError "The name of the remote machine must appear in one and only one file." 
    die "Third party file transfer disabled." 
fi

if [ "x$colonPresence1" = "x0" -a "x$colonPresence2" != "x0" ]; then
    # The following is a put: define first and second parameter of the put command
    transferType="put"
    remoteFile=${second_arg_file}
    remotePath=`echo $remoteFile | cut -d":" -f2`
    localFile=${first_arg_file}
fi

if [ "x$colonPresence1" != "x0" -a "x$colonPresence2" = "x0" ]; then
    # The following is a get: define first and second parameter of the get command
    transferType="get"
    remoteFile=${first_arg_file}
    remotePath=`echo $remoteFile | cut -d":" -f2`
    localFile=${second_arg_file}
fi

# If the remote path is not absolute, will not use the remote user directory
# as the std rcp. Here it is an error
slashPosRemPath=`echo $remotePath | awk '{print index($1,"/")}' -`
[ "x$slashPosRemPath" != "x1" ] && die "Error: the remote path must be absolute"

# If the local file has no absolute path, try to use $PWD to define it
slashPosLocFile=`echo $localFile | awk '{print index($1,"/")}' -`
if [ "x$slashPosLocFile" != "x1" ]; then
  localFile="${PWD}/$localFile"
fi

# Now define what the source and destination URL are. Note that here
# the format is not complete, yet.
if [ "x$transferType" = "xput" ]; then
    sourceURL=$localFile
    destURL=$remotePath
    sourceProtocol=$localProtocol
    destProtocol=$remoteProtocol
else
    sourceURL=$remotePath
    destURL=$localFile
    sourceProtocol=$remoteProtocol
    destProtocol=$localProtocol
fi

# Check name consistency.
# -First of all is there a file name specified?
#   gridftp cannot copy directories.
# -Is the second parameter a path to a directory or to a file with the same
#  file name?
#   To make things easier the local and remote file names must be the same:
#   in case of a "put", there is not an easy way with gridftp one can check
#   on the remote host if the last part of a path is a file name or a dir.
#   Here,  the file name is always appended to the second parameter path, if
#   the last part of the second parameter path is not the same file name.

sourceFileName=`echo $sourceURL | awk 'BEGIN{FS = "/"}{print $NF}' -`
destFileName=`echo $destURL | awk 'BEGIN{FS = "/"}{print $NF}' -`
[ "x$sourceFileName" = "x" ] && die "Error: you must provide a file name"

# The file name must appear at least (but not only) in the first par path
if [ "x$sourceFileName" != "x$destFileName" ]; then
    if [ "x$destURL" = "x" ]; then
	temp="${sourceFileName}"  		#in case destURL = "<machine>:"
    else
	temp="${destURL}/${sourceFileName}" # having // in the path is ok
    fi
    destURL=$temp
fi

# Parse machine name and user name
# Check if the user name is included in the path and check if it is sam
remoteUserMachine=`echo $remoteFile | cut -d":" -f1`
# Is there @ in remoteUserMachine ?
tstString=`echo $remoteUserMachine | awk '{print index($1,"@")}' -`

if [ "x$tstString" != "x0" ]; then
    remoteMachine=`echo $remoteUserMachine | cut -d"@" -f2`
else 
    remoteMachine=$remoteUserMachine
fi


if [ -z "${serverType}" ]; then
   serverType="${default_serverType}"
   echoInfo "Set server type to default: ${serverType}"
fi

# get server subject
# if serverSubject is not supplied 1st check environment
if [ -z "${serverSubject}" -a -n "${JIM_GRIDFTP_SERVER_SUBJECT}" ]; then
   serverSubject="${JIM_GRIDFTP_SERVER_SUBJECT}"
   echoInfo "Read server subject from environment: ${serverSubject}" 
fi

# next configuration
if [ -z "${serverSubject}" -a -n "${JIM_CONFIG_DIR}" ]; then
   echoInfo "Will read server subject for host $remoteMachine from configuration"
   serverSubject=`jim_config_manager_cmd.py gca -n jim_gridftp -p "/jim_gridftp_configuration/\
host[@name=\"${remoteMachine}\"]/${serverType}/certificate" -a subject -m FS`
   [ $? = 0 ] || serverSubject=""
fi

# guess server subject to be doe cert type subject
if [ -z "${serverSubject}" ];then
   serverSubject="/DC=org/DC=doegrids/OU=Services/CN=sam/${remoteMachine}"
   echoWarning "Guessing server subject as: $serverSubject"  
fi

# get port number
# if port number is not supplied 1st check environment
if [ -z "${port}" -a -n "${JIM_GRIDFTP_SERVER_PORT}" ]; then
   port="${JIM_GRIDFTP_SERVER_PORT}"
   echoInfo "Read port number from environment: ${port}"
fi

# next configuration
if [ -z "${port}" -a -n "${JIM_CONFIG_DIR}" ]; then
   echoInfo "Will read port number for host $remoteMachine from configuration"
   port=`jim_config_manager_cmd.py gca -n jim_gridftp -p "/jim_gridftp_configuration/\
host[@name=\"${remoteMachine}\"]/${serverType}/port" -a number -m FS`
   [ $? = 0 ] || port=""
fi

# no other choice then default
if [ -z "${port}" ];then
   port="${default_port}"
   echoInfo "Set port number to default: ${port}"
fi


#Finalize source and dest URL. Note that source and dest URL are absolute
#path here
if [ "x$transferType" = "xput" ]; then
  sourceURL="${sourceProtocol}://${localMachine}${sourceURL}"
  destURL="${destProtocol}://${remoteMachine}:${port}${destURL}"
  ssubject=${user_proxy_subject}
  dsubject=${serverSubject}
else
  sourceURL="${sourceProtocol}://${remoteMachine}:${port}${sourceURL}"
  destURL="${destProtocol}://${localMachine}${destURL}"
  ssubject=${serverSubject}
  dsubject=${user_proxy_subject}
fi

#####
##### Transfers the file
#####

#Compatibility with globus v2.0: option -nodcau
${GLOBUS_LOCATION}/bin/globus-url-copy \
 -nodcau \
 -ss "${ssubject}" -ds "${dsubject}" \
 ${sourceURL} ${destURL} 1>&1 2>&2


if [ "$?"  -ne "0" ]; then
        echoError "Configuration Read:"
        echoError "	Port           = ${port}"
        echoError "     Remote Machine = ${remoteMachine}"
        echoError "	Server Subject = ${serverSubject}"  
	echoError "	User Subject   = ${user_proxy_subject}"
	if [ -n "${JIM_CONFIG_DIR}" ]; then
	   echoError "jim_gridftp product Configuration is"
	   echo ""
	   jim_config_manager_cmd.py gcs -n jim_gridftp -m FS 1>&2
	   echo ""
	fi
	die "Failed to copy file ${sourceURL} to ${destURL}"
else
	echoInfo File copied successfully
fi
exit 0
