#!/bin/bash
#This script starts up the gridftp server as user sam. It relies on a
#previous installation of globus ftp. The gridftp client (i.e.
#grid-url-copy) talks to it with the option 
# -s '/O=Grid/O=Globus/OU=fnal.gov/CN=SAM Run II Sequential Access'
#In case of future need, with slight modification this can become the
#sam_bootstrap wrapper to check whether the server is still alive
#(see sam_bbtfp).

prog=`basename $0`
echoInfo()    { echo "$prog: `date "+%a %D %T"`: INFO   : $*" >&1; }
echoWarning() { echo "$prog: `date "+%a %D %T"`: WARNING: $*" >&2; }
echoError()   { echo "$prog: `date "+%a %D %T"`: ERROR  : $*" >&2; }
die () { echoError $*; exit 1; }
usage () { echo "$prog --server-type=[server-name]" >&2; exit 1; }
get_value () { echo $1 | sed "s?^--[a-z|A-Z|0-9|-]*=??g"; }

[ -z ${SAM_GSI_CONFIG_DIR} ] && die "Please, setup sam_gsi_config"
[ -z ${GLOBUS_LOCATION} ] && die "Please, set GLOBUS_LOCATION env var"
[ -z "${JIM_CONFIG_DIR}" ] && die "Please, setup jim_config"

## parse arguments
while [ -n "$1" ]; do
  case $1 in
  --server-type=*)
    server_type=`get_value "$1"`
    echo "$prog: Starting server type: ${server_type}"
    ;;

  *)
    usage
  esac
  shift
done
#get the server type
[ -n "$server_type" ] || usage


if [ -n "$X509_USER_PROXY" ];then 
   echoWarning "Found X509_USER_PROXY variable in the environment... will unset it"
   unset X509_USER_PROXY
fi

## read configuration from sam_gsi_confg | environment and jim_config

# this script is designed to work for the sam project
user="sam"  

# get user cert
if [ -n "${X509_USER_CERT}" ];then
   echoInfo "Read sam service certificate from environment"
else
   echoInfo "Reading sam service certificate from sam_gsi_config"
   X509_USER_CERT=`sam_gsi_read_config SAM_GSI_JIM_GRIDFTP_X509_CERT`
   [ $? = 0 ] || X509_USER_CERT=""
fi

[ -z "${X509_USER_CERT}" ] && die "Cannot find sam service certificate"
export X509_USER_CERT


# get user key
if [ -n "${X509_USER_KEY}" ];then
   echoInfo "Read sam service key from environment"
else
   echoInfo "Reading sam service key from sam_gsi_config"
   X509_USER_KEY=`sam_gsi_read_config SAM_GSI_JIM_GRIDFTP_X509_KEY`
   [ $? = 0 ] || X509_USER_KEY=""
fi
[ -z "${X509_USER_KEY}" ] && die "Cannot find user key"
export X509_USER_KEY


# get cert directory
if [ -n "${X509_CERT_DIR}" ];then 
   echoInfo "Read certificate directory from environment"   
else
   echoInfo "Reading X509_CERT_DIR  from sam_gsi_config" 
   X509_CERT_DIR=`sam_gsi_read_config SAM_GSI_JIM_GRIDFTP_CERT_DIR` 
   [ $? = 0 ] || X509_CERT_DIR=""
fi

if [ -z "${X509_CERT_DIR}" ];then
   [ -d "/etc/grid-security/certificates/" ] ||\
      die "Cannot find trusted CAs certificate directory"
   echoInfo "Setting X509_CERT_DIR to default: /etc/grid-security/certificates"
   X509_CERT_DIR="/etc/grid-security/certificates"
fi 
export X509_CERT_DIR 


# get gridmap file
if [ -n "${GRIDMAP}" ];then
   echoInfo "Read gridmap file from environment"
else
   echoInfo "Reading GRIDMAP from sam_gsi_config"
   GRIDMAP=`sam_gsi_read_config SAM_GSI_JIM_GRIDFTP_MAPFILE` 
   [ $? = 0 ] || GRIDMAP=""
fi

if [ -z "${GRIDMAP}" ]; then
   if [ -r "${SAM_GSI_CONFIG_DIR}/grid-security/grid-mapfile-jimsam" ]; then
      GRIDMAP="${SAM_GSI_CONFIG_DIR}/grid-security/grid-mapfile-jimsam"
   elif [ -r "/etc/grid-security/grid-mapfile" ];then
      GRIDMAP="/etc/grid-security/grid-mapfile"   
   else
      die "Cannot determine GRIDMAP file location"
   fi
   echoInfo "Set gridmap file to: ${GRIDMAP}"
fi

export GRIDMAP


# read the port number from the configuration
echoInfo "Reading port from jim_gridftp configuration"
stdGridFTPPort=""
host=`hostname -f`
if [ $? != 0 ];then
   echoWarning "Cannot execute hostname -f"
else
   stdGridFTPPort=`jim_config_manager_cmd.py gca -n jim_gridftp -p\
   "/jim_gridftp_configuration/host[@name=\"${host}\"]/${server_type}/port" -a number`
   [ $? = 0 ] || stdGridFTPPort=""
fi


if [ -z "${stdGridFTPPort}" ];then
   echoWarning "Using default port for running ${server_type}"
   case $server_type in
      data_server)
         stdGridFTPPort=4568
	 ;;
      head_server)
         stdGridFTPPort=4569
	 ;;	 
      *)
         die "Unknown server_type: ${server_type}"
   esac 
fi


[ "`whoami`" != "$user" ] && die "This script accepts $user as the only user allowed to startup the daemon."
echoInfo "Forcing permissions to no sam write access to prevent user tinkering." 
echoInfo "If gridmap not owned by sam then also safe."
chmod a-w ${GRIDMAP}

echoInfo "Server Configuration:"
echoInfo "      Server Type     : $server_type" 
echoInfo "      User Certificate: $X509_USER_CERT"
echoInfo "      User key        : $X509_USER_KEY"
echoInfo "      Cert Directory  : $X509_CERT_DIR"    
echoInfo "      Grid Mapfile    : $GRIDMAP"
echoInfo "      Port            : $stdGridFTPPort"   

echoInfo "Starting Server ..."

# -S  stand alone mode
# -p  port number
# -o  file transfers logged to xferlog: cannot change log file; it is $GLOBUS_LOCATION/var/log/xferlog
# -a  allow (USE -A instead )
# -Z  Globus logging
# -d  writes logs to /var/log/messages 
# -W don`t log user ogons in /var/log/wtmp, no permission as sam.
# Running as a non-root user:
#
# -Q  Don't use pid files since you don't have write access to them ( /var/run )
# -W  Don't record logins in wtmp (you don't have write permissions)
#
#Option description at http://www.wu-ftpd.org/man/ftpd.html


# the -A flag is important otherwise in.ftpd tries to read /etc/ftpacsess file which 
# maynot have the right permission
cmd="${GLOBUS_LOCATION}/sbin/in.ftpd -o -s -A -Q -W -p ${stdGridFTPPort} -G ${GLOBUS_LOCATION}"
# echo "$prog: EVAL: $cmd"
eval $cmd &
status=$?

if [ "x$status" = "x0" ]; then
  pid=`${JIM_GRIDFTP_DIR}/ups/echo_jim_gridftp_pid.sh`
  echoInfo "Started in.ftpd daemon with pid ${pid} as user ${user}"
else
  die "Cannot start in.ftpd daemon."
fi

while :
do
    pid=`${JIM_GRIDFTP_DIR}/ups/echo_jim_gridftp_pid.sh ${stdGridFTPPort}`

    if [ "x${pid}" = "x" ]; then
        break
    fi
    echoInfo "jim_gridftp server ($server_type) process still alive"
    sleep 3600
done
echoInfo "jim_gridftp server ($server_type) process died"


exit $status
