#!/usr/bin/env python
################################################################################
# Project   : SAM
# Package :
# $Id: RecocertImplementation.py,v 1.14 2007/02/26 21:55:55 parag Exp $
#
# This work is part of a development project, called SAM, which consists of a
# number of coordinated packages each named sam_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
#
# This file is used to run a project over a setup file configured
# by the user.  Please see project_setup.py for more information.
################################################################################

import os
import sys
import time
import random
import traceback
import string
import commands


import Fetcher
import MergeFlow
import MetadataHelper

import SAMDbServer
from Sam import sam
from samProject import Printers
from SamStruct.NameOrIdList import NameOrIdList
from SamStruct.SamSize import SamSize
from SamStruct.SamTime import SamTime
from SamStruct.ApplicationFamily import ApplicationFamily

from MetadataHelper import SamMDCache
from MetadataHelper import _metaAttributes
import jim_sam_storage_negotiator

class RecocertImplementation:

    def __init__(self, argDict):
        self._print = Printers("recocert")
        try:
            self._print.info("Initializing Recocert")
            self.doInit(argDict)
        except Exception, ex:
            traceback.print_exc(file=sys.stderr)
            self.updateProgress("Top level exception: " + str(ex))
            raise ex
    # __init__

    def doInit(self, argDict):
        if argDict.has_key('standAloneRun'):
            self._standAloneRun = argDict.get('standAloneRun')
        else:
            self._standAloneRun = 0
        self._dataTier = 'root-histogram'
        self._projectVersion = argDict.get('appversion')
        self._appFamily = 'certification'
        self._appName = 'recocert'
        self._appVersion = self._projectVersion
        self._jobFileDsId   = argDict.get('jobFilesDatasetId')
        self._jobFileDsName = argDict.get('jobFilesDatasetName')
        self._mergedFiles = argDict.get('mergedFiles')
        self._inFiles = self.getInputFileList()
        self._durableLocation = ""
        self._facilityname  = argDict.get('facilityname')
        self._project = self._consumerId = self._processId = None
        if argDict.has_key('durableLocation'):
            self._durableLocation = argDict.get('durableLocation')
      
        if argDict.has_key('debugLevel'):
            self._debug = int(argDict.get('debugLevel'))
        else:
            self._debug = 0

        self._commonargs = {
                                'project'    : argDict.get('project'),
                                'timeout'    : argDict.get('timeout'),
                                'station'    : argDict.get('station'),
                                'group'      : argDict.get('group'),
                           }

        if ( self._jobFileDsId ):
            self._jobFileDsId = int(self._jobFileDsId)

        # Attach the test attributes if required
        # NOTE: We only look at one file. The assumption is that all the files
        #       will have same application and we do not mix the test and 
        #       production versions.

        for file in self._inFiles.keys():
            fileMD = _metaAttributes.get(file)
            applicationFamily = fileMD.getApplicationFamily()
            if (applicationFamily.getAppName() == 'd0reco-test'):
                self._appFamily  = '=certification'
                self._appName    = 'recocert-test'
                self._appVersion = '%s-test' % self._projectVersion
            break

        # Get the settings for staging input files
        self._incomingNode = jim_sam_storage_negotiator.jim_sam_incoming_storage_negotiator(self._commonargs['station'], "dzero_reco_merge")
        self._commonargs['nodeName'] = self._incomingNode[0]
        if ( self._incomingNode[1] ):
            os.environ["SAM_FCP_QUEUE_NAME"] = self._incomingNode[1]

        # Get the settings for staging output files
        self._outgoingNode = jim_sam_storage_negotiator.jim_sam_outgoing_storage_negotiator(self._commonargs['station'], "fss_stager")

        # Start project/consumer/process to get a processId
        # We do not have any input dataset, use jobFilesDs instead as hack
        try:
            if not self._commonargs['project']:
                self._commonargs['project'] = "Recocert_%s" % os.environ.get('JIM_CLUSTER_JID')
            projectName = "%s.%f" % (self._commonargs['project'], random.random()) 
            self._print.info("Starting project with name %s ..." % projectName)
            self._project = sam.startProject(
                                group=self._commonargs['group'],
                                station=self._commonargs['station'],
                                project=projectName,
                                definitionName=self._jobFileDsName)
            self._print.info("Starting project with name %s ... Done" % projectName)
            self._print.info(self._project)

            self._print.info("Establishing consumer for project %s ..." % self._project.getProjectName())
            self._consumerId = sam.establishConsumer(
                                group=self._commonargs['group'],
                                station=self._commonargs['station'],
                                project=self._project.getProjectName(), 
                                appFamily=self._appFamily, 
                                appName=self._appName, 
                                appVersion=self._appVersion)
            self._print.info("Establishing consumer for project %s ... Done" % self._project.getProjectName())
            self._print.info(self._consumerId)

            self._print.info("Establishing process for project %s ..." % self._project.getProjectName())
            self._processId = sam.establishProcess(
                                station=self._commonargs['station'],
                                project=self._project.getProjectName(), 
                                cid=self._consumerId,
                                nodeName=self._commonargs['nodeName'])

            self._print.info("Establishing process for project %s ... Done" % self._project.getProjectName())
            self._print.info(self._processId)

        except Exception, ex:
            traceback.print_exc(file=sys.stderr)
            self.stopSamProcesses()
            raise ex

        cwd = os.getcwd()
        d0runjob_jobname = "%s-recocert" % os.environ['JIM_GLOBAL_JID']
        d0runjob_jobdir = "%s/%s-1" % (cwd, d0runjob_jobname)
        outputDir = "%s/root-histograms" % (cwd)
        os.system("rm -rf %s" % outputDir)
        os.makedirs(outputDir)

        self._outFiles = self.getOutputFileList(outputDir)

        try:
            # Fetch the files you want to run recocert on
            if (len(self._inFiles) == 0):
                raise Exception("There are no thumbnails to run recocert")

            # Stage Job Files
            if self._jobFileDsId or self._jobFileDsName :
                self.stageJobFiles()

            for fileName in self._inFiles.keys():
                self._print.info("Processing file %s" % fileName)
                self.runRecocert(fileName, outputDir,
                                 d0runjob_jobname, d0runjob_jobdir)

        except:
            traceback.print_exc(file=sys.stderr)
            self.packCustomOutput(cwd)
            self.stopSamProcesses()
            raise

        self.packCustomOutput(cwd)
        self.stopSamProcesses()
    # doInit


    def stopSamProcesses(self):
        try:
            sam.endProcess(station=self._commonargs['station'],
                           project=self._project.getProjectName(),
                           processId=self._processId)
            sam.stopProject(station=self._commonargs['station'],
                            project=self._project.getProjectName(),
                            force=1)
        except Exception, ex:
            traceback.print_exc(file=sys.stderr)
            self._print.warning("Unable to stop SAM process or project")
    # stopSamProcesses


    def packCustomOutput(self, cwd):
        if (self._standAloneRun):
            shipOutput = """cd %s; ls -laR %s* > ./created_files.txt; tar cvfz $OUTPUT_FILE `find . -name *log` ./created_files.txt `find . -name *core*` `find . -name events.copy*` `find . -name *macro*` `find . -name import_*py`""" % (cwd, cwd)
            os.system(shipOutput)
    # packCustomOutput


    def getInputFileList(self):
        files = {}
        for file in self._mergedFiles.split(','):
            filename = file.split('/')[-1]
            if filename.startswith("recoT"):
                files[filename] = file
        return files
    # getFileList


    def getOutputFileList(self, outputDir):
        outputFileList = {}
        for file in self._inFiles.keys():
            outputFileList[file] = "%s/cert_%s.root" % (outputDir, file)
        return outputFileList
    # getFileList


    def getD0runjobDir(self, path):
        import os
        list = os.listdir(path)
        pattern = "d0runjob"

        for i in list:
           if i.startswith(pattern):
              return i

        raise Exception("Unable to get the D0_RUNJOB_DIR")
    # getD0runjobDir


    def untar(self, pattern):
        cmd = "tar xzf %s && rm -f %s" % (pattern,pattern)
        status,stdouterr = commands.getstatusoutput(cmd)
        if status != 0:
            raise Exception("untar failed %s" % stdouterr)
    # untar

    def stageJobFiles(self):
        cwd = os.getcwd()
        if (self._standAloneRun):
            try:
                stream = Fetcher.FetchStream(\
                    cwd,\
                    Fetcher.getFileIdList(self._jobFileDsId, \
                                          self._jobFileDsName),\
                    self._commonargs)
                while 1:
                    chunk = stream.fetch(5)
                    if ( chunk == None ):
                        break
                self.updateProgress("Finished staging job files")
            except Exception,ex:
                self._print.error("Failed to stage job files")
                self.stopSamProcesses()
                raise ex
            try:
                self._print.info("Unpacking d0 code")
                self.untar("d0_*.tar.gz")
                self._print.info("Unpacking d0runjob")
                self.untar("d0runjob*.tar.gz")
            except Exception,ex :
                self._print.error("Failed to unpack job files")
                self.stopSamProcesses()
                raise ex
        else:
            self._print.info("Running recocert after merging. Staging job files not required yet. Just set the SRT_DIST and D0_RUNJOB_DIR environment variables")
        self._print.info("Setting up d0runjob")
        os.environ['D0_RUNJOB_DIR'] = "%s/%s" % (cwd, self.getD0runjobDir(cwd))
        self._print.info("Setting up d0 code")
        os.environ['SRT_DIST'] = "%s/RTE_runjob" % cwd
    # stageJobFiles


    def runRecocert(self,fileName,outputDir,d0runjob_jobname,d0runjob_jobdir):
        self._print.info("Running recocert on %s" % fileName)
        self.updateProgress("Running recocert on %s" % fileName) 

        # Generate d0runjob Macro
        macroFile = self.createMacro(fileName, outputDir,
                                     self._projectVersion, d0runjob_jobname)

        # Generate Initial Metadata for the histogram
        initialMD = self.generateInitialMD(fileName)

        self.updateProgress("Finished setup, launching d0runjob for recocert")

        # Run the macro file with d0runjob
        self.d0runjob(macroFile)

        # Do any consistency checks as required
        outFiles = os.listdir(outputDir)
        self._print.info("Contents of dir %s: %s" % (outputDir, str(outFiles)))

        # Generate the final Metadata
        mdFileName = self.generateMD(initialMD, outputDir,
                         MetadataHelper._samMDCache.get(initialMD['parents'][0]))

        # Store the file to SAM
        self.storeOne(initialMD, outputDir, mdFileName)
    # runRecocert


    def createMacro(self, fileName, outputDir, d0release, d0runjob_jobname):

        self._print.info("Generating recocert macro")
        self.updateProgress("Generating recocert macro") 

        working_dir = os.getcwd()
        srt_dist = os.environ['SRT_DIST']
        srt_dist_dir = srt_dist + "/releases/" + d0release
        gjid = os.environ['JIM_GLOBAL_JID']

        outputFile = self._outFiles[fileName]

        macro="""
import os

locald0 = os.path.expandvars("%s")

from d0rjcore.d0userinterface import createUserInterface
from d0rjcore.d0globals import D0Globals

d0rj = createUserInterface(
    JobName = "%s",
    JobCacheArea = "%s",
    OptionalMonitors = ['jim'],
)
d0rj.clearLogVeto(D0Globals.LogDbgLo)
d0rj.clearLogVeto(D0Globals.LogDbgMed)
d0rj.clearLogVeto(D0Globals.LogDbgHi)


fileIter = d0rj.attachConfigurator("FileIterator")
fileIter["InputFiles"] = %s
fileIter["IncrementSize"] = 0

d0recocert = d0rj.attachExecutable("D0RecoCert")
d0recocert["InputFiles"] = "::FileIterator:Output"
d0recocert["OutputFile"] = "%s"
d0recocert["ProjectVersion"] = "%s"
d0recocert["DistroDirectory"] = "%s"
#d0recocert["FrameworkRCPName"] = "runD0ChunkAnalyze_TMB.rcp"
d0recocert["FrameworkRCPName"] = "runD0ChunkAnalyze_TMB_metdq.rcp"
#d0recocert["FrameworkRCPName"] = "runD0ChunkAnalyze_TMB_noAA_noclusters.rcp"
d0recocert["RcpDirectory"] = locald0

# Register general job manager with BashJobOperator in order to produce
# the required job and task scipts
bashjob = d0rj.attachConfigurator("BashJobOperator")
bashjob.registerManager(d0rj.getManager())

prov = d0rj.attachConfigurator("ProvenanceProcessor")
prov.registerConfigurator(fileIter)
prov.registerConfigurator(d0recocert)

run = d0rj.attachConfigurator("RunInteractive")
run.registerManager(d0rj.getManager())

for i in range(0, 1):
  d0rj.run()
""" % (srt_dist_dir, d0runjob_jobname, working_dir, [self._inFiles[fileName]], outputFile, d0release, srt_dist)

        macroFile ="recocert.macro"
        try:
            f = open(macroFile, 'w')
            f.write(macro)
            f.close()
        except Exception, ex:
            self._print.error("Failed to write macro file")
            self.stopSamProcesses()
            raise ex

        print "Macro dumped in file %s" % macroFile
        print "------------------------------------"
        print "         Generated Macro            "
        print "------------------------------------"
        print "%s" % macro
        print "------------------------------------"
        sys.stdout.flush()
        return macroFile
    # createMacro


    def d0runjob(self,macroFile):
        # Setup the environment for d0runjob
        os.environ['PYTHONPATH'] = os.environ['PYTHONPATH'] + ":" + os.environ['D0_RUNJOB_DIR'] + "/Python"
        os.environ['SHAHKAR_FACTORY_EXT'] = os.environ['D0_RUNJOB_DIR'] + "/xml/d0factory.xml"

        try :
            if not os.environ['D0_RUNJOB_DIR']:
               raise Exception("Need D0_RUNJOB_DIR env variable set")

            print "Invoking d0runjob to generate root-histograms files"
            cmd = "python %s" %(macroFile)
            sys.stdout.flush()
            print "==============================================\n"
            print os.environ
            print "==============================================\n"
            print "Executing %s" % (cmd)
            status = os.system(cmd)
            print "Exit code = %d" % status
            print "\n=============================================="

            if status != 0:
                raise Exception("d0runjob returned an invalid status code")

        except Exception,ex:
             self._print.error("d0runjob failed !")
             self.stopSamProcesses()
             raise ex
   # d0runjob


    def generateInitialMD(self, fileName):
        outFile = "cert_%s.root" % fileName
        inputFileMD = _metaAttributes.get(fileName)

        initialMD = inputFileMD
        initialMD['filename'] = outFile
        initialMD['firstEvent'] = inputFileMD.getFirstEvent()
        initialMD['lastEvent'] = inputFileMD.getLastEvent()
        initialMD['totalevents'] = inputFileMD.getEventCount()
        initialMD['parents'] = [fileName]
        initialMD['dataTier'] = self._dataTier
        self._print.info("==================================================")
        self._print.info("Initial metadata generated")
        self._print.info("==================================================")
        self._print.info(initialMD)
        self._print.info("==================================================")
        return initialMD
    # generateMetadata


    def generateMD(self, initialMD, outputDir, parentMD):
        fileName = initialMD['fileName']

        fileStruct = None
        try:
            fileStruct = os.stat(os.path.join(outputDir, fileName))
        except OSError,ex:
            self.updateProgress("job generated no output file "+fileName)
            self.stopSamProcesses()
            raise ex

        fileSize   = fileStruct[6]/1024
        fileCreateTime = fileStruct[9]
        fileModifiedTime = fileStruct[8]

        header = """from SamFile.SamDataFile import DerivedDetectorFile
from SamFile.SamDataFile import ImportedSimulatedFile
from SamStruct.SamSize import SamSize
from SamStruct.CRC import CRC
from SamStruct.NameOrIdList import NameOrIdList
from SamUtility.CaseInsensitiveDictionary import CaseInsensitiveDictionary
from SamStruct.SamTime import SamTime
from SamStruct.ApplicationFamily import ApplicationFamily
from SamStruct.Params import Params
from SamStruct.RunDescriptorList import RunDescriptorList
from SamStruct.RunDescriptor import RunDescriptor
from SamStruct.ParamValue import ParamValue
from SamStruct.DataType import DataType
from SamStruct.LumBlockRangeList import LumBlockRangeList
from SamStruct.LumBlockRange import LumBlockRange
from SamStruct.SamList import SamList
import SAM
"""

        newMDStr= header + "newMD = "+str(parentMD)
        exec newMDStr

        newMD["fileName"] = fileName
        newMD["parents"] = NameOrIdList(initialMD["parents"])
        newMD['firstEvent'] = initialMD['firstEvent']
        newMD['lastEvent'] = initialMD['lastEvent']
        newMD["eventCount"] = initialMD['totalevents']
        newMD["fileSize"] = SamSize(str(fileSize)+"K")
        newMD["startTime"] = SamTime(fileCreateTime)
        newMD["endTime"] = SamTime(fileModifiedTime)
        newMD["dataTier"] = initialMD['dataTier']
        newMD["applicationFamily"] = ApplicationFamily(
                                         appFamily=self._appFamily,
                                         appName=self._appName, 
                                         appVersion=self._appVersion)
        newMD["processId"] = self._processId

        if (newMD.get('params')):
            if (newMD.get('params').has_key('thumbnail')):
                del(newMD.get('params')['thumbnail'])
            if (newMD.get('params').has_key('global')):
                if (newMD.get('params').get('global').has_key('runjobversion')):
                    del(newMD.get('params').get('global')['runjobversion'])

        if ( self._facilityname ):
            if ( newMD.get('params') ):
                if ( newMD.get('params').get('Global') ):
                    newMD.get('params').get('Global')['facilityname'] = self._facilityname
                else:
                    newMD.get('params')['Global'] = { 'facilityname' : self._facilityname }
            else:
                newMD['params'] = Params({'Global' : {'facilityname': self._facilityname } } )

        newMD["crc"] = sam.calculateFileCrc(path=os.path.join(outputDir,fileName))

        self._print.info("SAM Generated MD :"+str(newMD))
  
        self._print.info("==================================================")
        self._print.info("           Generated Metadata Info")
        self._print.info("==================================================")
        mdFileName = "%s/import_%s.py" % (outputDir, fileName)
        print "Metadata file: %s" % mdFileName
        self._print.info("--------------------------------------------------")
        try :
            mdFile = open(mdFileName, 'w')
            mdFile.write(header + "t=" + str(newMD))
            mdFile.close()
        except Exception, ex :
            self._print.error("Failed to write metadata file %s" % mdFileName)
            self.stopSamProcesses()
            raise ex
  
        print "Generated Metadata file reads:\n"
        try:
            f = open(mdFileName, 'r')
            for line in f.readlines():
                print line
            f.close()
        except:
            self._print.error("Failed to read metadata file %s" % mdFileName)
        self._print.info("==================================================")
        return mdFileName
    # generateMetadata


    def storeOne(self, basicMD, outDir, mdFileName):
        self._print.info("Trying to store the root-histogram file with metadata in %s " % mdFileName )

        fileName = basicMD['filename']

        if ( self._debug ):
            self._print.info("File "+fileName+" will not be stored in debug mode")
            # Additional info
            print "------------------------------------------------------------"
            print " FOLLOWING COMMANDS WERE UNEXECUTED AS THIS IS TEST RUN"
            print "------------------------------------------------------------"
            print " Printing from location %s " % (os.getcwd())
            print "Using outgoing node : %s and fcp q %s " % ( self._outgoingNode[0] , self._outgoingNode[1] )
            print "------------------------------------------------------------"
            sys.stdout.flush()
            return

        self._print.info("Using outgoing node : %s and fcp q %s " % ( self._outgoingNode[0], self._outgoingNode[1] ) )

        try:
            filepath = os.path.join(outDir,fileName)

            if self._outgoingNode[1]:
                os.environ['SAM_FCP_QUEUE_NAME'] = self._outgoingNode[1]

            if self._outgoingNode[0]:
                transferRequestId = sam.requestFileTransfer(
                                        sourceFile=filepath, \
                                        destinationPath=self._durableLocation,\
                                        declareFile=1,\
                                        temporaryTransferLocation=self._outgoingNode[0],\
                                        waitForCompletion=1,\
                                        retryMaxCount=3,\
                                        descriptionFile=mdFileName)
            else:
                transferRequestId = sam.requestFileTransfer(
                                        sourceFile=filepath, \
                                        destinationPath=self._durableLocation,\
                                        declareFile=1,\
                                        waitForCompletion=1,\
                                        retryMaxCount=3,\
                                        descriptionFile=mdFileName)
        except Exception,ex:
            traceback.print_exc(file=sys.stderr)
            self._print.error("Failed to store file with metadata in %s" % \
                              mdFileName)
            raise "Failed to store file with metadata in %s" % mdFileName

        sys.stderr.flush()
        sys.stdout.flush()

        locList = MetadataHelper.getFileLocation(fileName)
        if ( not len(locList) ):
           raise Exception("File store finished but no location has been registered for "+fileName)

        self.updateStateInDB("stored",fileName)
    # storeOne


    def updateStateInDB(self, state, filename):
        updateStr = '<output_file name=\"%s\" time=\"%s\" time_epoch=\"%s\" state=\"%s\"/>' % (filename, time.strftime("%b %d %Y %H:%M:%S %Z", time.localtime()), str(int(time.time())), state)
        self.updateDB(updateStr)
    # updateStateInDB


    def updateProgress(self, progressStr):
        updateStr = """'<progress created=\"%s\" time_epoch=\"%s\">%s</progress>'""" % ( time.strftime("%b %d %Y %H:%M:%S %Z", time.localtime()) , str(int(time.time())), progressStr )
        self.updateDB(updateStr)
    # updateProgress

    def updateDB(self, updateStr):
        """Update the JIM XML DB"""
        try:
            gjid = os.environ['JIM_GLOBAL_JID']
            cjid = os.environ['JIM_CLUSTER_JID']
            ljid = os.environ['JIM_LOCAL_JID']
            url  = os.environ['XMLDB_SERVER_URL']
        except Exception, ex:
            traceback.print_exc(file=sys.stderr)
            self._print.warning("%s env variable missing" % ex)
            self._print.warning("Cannot update XML database")
            sys.stderr.flush()
            return

        try:
            self._print.info("Updating XML database")

            jobCollection = "/db/jobs/globalJobs1"
            cmd = """xmldb_cmd id -c %s -n %s_%s -x '/local_job[@Id="%s"]' -url %s -xs '%s'""" % (jobCollection, gjid, ljid, ljid, url, updateStr)
            self._print.info("Executing: %s" % cmd)
            os.system(cmd)

            #from xmldb_client import idmgr
            #jobCollection = "/db/jobs/%s" % (gjid)
            #db = idmgr.idmgr(url, jobCollection)
            #db.UpdateLocal("append", gjid, cjid, ljid, updateStr)
        except Exception, ex:
            traceback.print_exc(file=sys.stderr)
            self._print.warning(ex)
            self._print.warning("Could not Update XML Database...will continue")
            sys.stderr.flush()
   # updateDB

# class RecocertImplementation
