#!/usr/bin/env python

################################################################################
# Project   : SAM
# Package : sam_user_api
# $Id: GenericRunProjectScript.py,v 1.8 2006/04/10 22:57:54 illingwo Exp $
#
# This work is part of a development project, called SAM, which consists of a
# number of coordinated packages each named sam_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
#
# This file is used to run a project over a setup file configured
# by the user.  Please see project_setup.py for more information.
#
################################################################################

import sys
import string
import os
import signal

from Sam import sam

from SamUserApi_UserImport import userImport
import SAM

DEFAULT_CPU_PER_EVENT = "1s"
DEFAULT_FILE_RELEASE_STATUS = "ok"

################################################################################

class GenericRunProjectScript:
    def __init__(self, userFileReadyScript):

        #
        # things that we will set on the user's behalf:
        self._projectName = ""
        self._cid = -1
        self._cpid = -1

        # these are the parameters that may be set in the GenericRunProjectScript
        self.USER_PARM_LIST = ['sam_project',
                               'sam_station',
                               'project_definition',
                               'dataset_version',
                               'appname',
                               'appfamily',    # new, v7_3_1, to pass appfamily as well as appName, version
                               'app_family',   # very very old, means appname in ancient project files
                               'version',
                               'group',
                               'max_file_amt',
                               'cpu_per_event',
                               'give_up',
                               'erralertflag',
                               'timeout',
                               'file_ready' ,
                               'begin_project' ,
                               'end_project',
                               'abort_project',
                               'persistent_process',
                               'detached']
        # initialize these:
        self.USER_PARM_DEFAULTS = {}
        for x in self.USER_PARM_LIST:
            if (x == "timeout"):
                self.USER_PARM_DEFAULTS[x] = 0
            elif (x == "cpu_per_event"):
                self.USER_PARM_DEFAULTS[x] = DEFAULT_CPU_PER_EVENT
            else:
                self.USER_PARM_DEFAULTS[x] = ""

        import types
        # override by reading input file (which imports a USER_PARM name space)
        if (type(userFileReadyScript) == types.StringType):
            if string.find(userFileReadyScript, '.py') == -1 :
               userFileReadyScript = "%s.py" % userFileReadyScript

            self.USER_PARM = self.read(userFileReadyScript)
        else:
            self.USER_PARM = userFileReadyScript

        # and for each of the elements we're looking for, make sure
        # that the attribute is set (i.e., fill in with defaults if
        # the user didn't set it):
        for attr in self.USER_PARM_LIST:
            if (not hasattr(self.USER_PARM, attr)):
                self.USER_PARM.__dict__[attr] = self.USER_PARM_DEFAULTS[attr]

    def __getattr__(self, attr):
        if (attr in self.USER_PARM_LIST):
            return self.USER_PARM.__dict__[attr]
        else:
            return self.__dict__[attr]

    ############################################################################
        
    def read(self,userFileReadyScript):
        # import the user's script
        try:
            USER_PARM = userImport(userFileReadyScript)
            importOk = 1
        except:
            sys.stderr.write("WARNING: an exception was generated while importing the project file:\n")
            sys.stderr.write("       %s %s\n" % (sys.exc_info()[0], sys.exc_info()[1]))
            importOk = 0
            
        try:
            # backwards compatibility for very old project files:
            if (not hasattr(USER_PARM, 'appname')):
                try:
                    USER_PARM.appname = USER_PARM.app_family
                except AttributeError:
                    USER_PARM.appname = ''
                    
            if (not hasattr(USER_PARM, 'dataset_version')):
                try:
                    USER_PARM.dataset_version = USER_PARM.snapshot_version
                except AttributeError:
                    USER_PARM.dataset_version = ''

            if (not importOk) :
                sys.stderr.write("CONTINUING after IMPORT exception...\n")
        except NameError:
            sys.stderr.write("ABORTING after IMPORT exception.\n")
            sys.exit(1)

        # backwards compatibility for project files that specify appName, version but not appFamily
        if (not hasattr(USER_PARM, 'appfamily')):
            USER_PARM.appfamily = SAM.SamDefault_Name_Unknown

        # backwards compatibility for project files without begin_project, end_project:
        if (not hasattr(USER_PARM, 'begin_project')):
            USER_PARM.begin_project = self.dummy_begin
        if (not hasattr(USER_PARM, 'end_project')):
            USER_PARM.end_project = self.dummy_end
        if (not hasattr(USER_PARM, 'abort_project')):
            USER_PARM.abort_project = self.dummy_abort
            
        # give the user access to useful utilities that we wrote on their behalf:
        if (not hasattr(USER_PARM, 'getStationName')):
            USER_PARM.getStationName = self.getStationName
        if (not hasattr(USER_PARM, 'getProjectName')):
            USER_PARM.getProjectName = self.getProjectName
        if (not hasattr(USER_PARM, 'getConsumerId')):
            USER_PARM.getConsumerId = self.getConsumerId
        if (not hasattr(USER_PARM, 'getConsumerProcessId')):
            USER_PARM.getConsumerProcessId = self.getConsumerProcessId
        if (not hasattr(USER_PARM, 'dumpProject')):
            USER_PARM.dumpProject = self.dumpProject
        if (not hasattr(USER_PARM, 'dumpConsumer')):
            USER_PARM.dumpConsumer = self.dumpConsumer

        # return the USER_PARM name space so that others can access it!
        return USER_PARM

    ############################################################################
   
    def dummy_begin(self):
        # The user has not provided a begin_project hook in the script
        return

    def dummy_end(self):
        # The user has not provided a end_project hook in the script
        return

    def dummy_abort(self):
        # The user has not provided an abort_project hook in the script
        return

    ############################################################################
    # items that the user may wish to know, that are not set by the user
    # in the project file:

    def setStationName(self, stationName):
        self._stationName = stationName

    def getStationName(self):
        return self._stationName
    
    def setProjectName(self, projectName):
        self._projectName = projectName

    def getProjectName(self):
        return self._projectName

    def setConsumerId(self, cid):
        self._cid = cid

    def getConsumerId(self):
        return self._cid

    def setConsumerProcessId(self, cpid):
        self._cpid = cpid

    def getConsumerProcessId(self):
        return self._cpid

    def dumpProject(self):
        print sam.dumpProject(station=self.getStationName(), project=self.getProjectName())

    def dumpConsumer(self):
        print sam.dumpConsumer(station=self.getStationName(),
                               project=self.getProjectName(),
                               cid=self.getConsumerId())

    
################################################################################
################################################################################

class FileReceiverAdapter:
    def __init__(self,station,project,cid,erralertflag,timeout):
        self._station = station
        self._project = project
        self._cid = cid
        self._erralertflag = erralertflag
        self._timeout = timeout

    def getCpid(self):
        pass
    def getNextFile(self):
        pass
    def terminate(self):
        pass

class TransientFileReceiverAdapter(FileReceiverAdapter):
    def __init__(self,station,project,cid,erralertflag,timeout):
        FileReceiverAdapter.__init__(self,station,project,cid,erralertflag,timeout)

        self._cpid = sam.establishProcess(station=station, project=project, cid=cid)

    def getCpid(self):
        return self._cpid

    def getNextFile(self):
        return  sam.getNextFile(station=self._station,
                                project=self._project,
                                cpid=self._cpid,
                                erralertflag=self._erralertflag,
                                timeout=self._timeout)

import os
class PersistentFileReceiverAdapter(FileReceiverAdapter):
    def __init__(self,station,project,cid,erralertflag,timeout):
        FileReceiverAdapter.__init__(self,station,project,cid,erralertflag,timeout) 
        comm_pipe = os.pipe()
        self._pid = os.fork()
        self._cpid = 0
        if self._pid == 0:
           os.close(comm_pipe[0])
          # sam.establishPersistentProcess(station=station, project=project, cid=cid, outputfile=comm_pipe[1])
           cmd = "sam create persistent process --station=%s --project=%s --cid=%d --outputfile=%d" % \
                 (station, project , cid , comm_pipe[1])
           os.execvp("/bin/sh",["/bin/sh","-c",cmd])
           import sys
           sys.exit(0)

        os.close(comm_pipe[1])   

        self._f = os.fdopen(comm_pipe[0],"rt")
   
    def getCpid(self):
        if self._cpid == 0:
           buffed_line = self._read()
           self._cpid = self._getCpid(buffed_line)
           if string.find(buffed_line,"ESTABLISHED") == -1:
              raise "ESTABLISHED token is expected first"

        return self._cpid

    def getNextFile(self):
        self.getCpid() 
        return self._getFileName(self._read())

    def terminate(self):
        import signal,time
        time.sleep(1)
        try:
            os.kill(self._pid,signal.SIGTERM)
        except:
            pass
        
    def _read(self):
       l = self._f.readline()
       return string.replace(l,"\n","")
 
    def _validate(self,l):
       t = string.split(l," ",2)
       if len(t) < 3:
          raise l +" : parse error, should be CPID <num> <filepath>"
       return t

    def _getCpid(self,l):
       t = self._validate(l)
       return int(t[1])

    def _getFileName(self,l):
       t = self._validate(l)
       return t[2]
           
        

################################################################################
################################################################################

class RunProject:

    def __init__(self, userFileReadyScript):
        self.userFileReadyScript = userFileReadyScript

    ############################################################################

    def ABNORMAL_EXIT(self, signum, frame):
        print(">>>>>>>> PROJECT ABORTED <<<<<<<<")
        self.userProject.USER_PARM.abort_project()
        sys.exit(1)
        
    ############################################################################

    def run(self):
        goodExitStatus = 1
        badExitStatus = 0
        # set up signal handlers so that we can exit cleanly if killed or
        # CTRL/C or something:
        signal.signal(signal.SIGTERM, self.ABNORMAL_EXIT)
        signal.signal(signal.SIGINT, self.ABNORMAL_EXIT)

        try:
            self.doTheDeed()
            returnValue = goodExitStatus
        # handle CTRL/C the same way as KILL:
        except KeyboardInterrupt:
            self.ABNORMAL_EXIT(signal.SIGINT, None)
            returnValue = badExitStatus
        return returnValue
            
    def doTheDeed(self):
        # get the user's project file parameters
        userProject = GenericRunProjectScript(self.userFileReadyScript)

        # we ASSUME that the project has already been started;
        #  hence we can use the environment to determine the project/station
        
        sam_station = ""
        try:
            sam_station = userProject.sam_station
        except:
            pass

        if not len(sam_station):
           sam_station = os.environ['SAM_STATION']

        sam_project = ""
        try:
            sam_project = userProject.sam_project
        except:
            pass
      
        if not len(sam_project):
            sam_project = os.environ['SAM_PROJECT']

        # other info from the user's project file information:
        appname       = userProject.appname
        appfamily     = userProject.appfamily
        appversion    = userProject.version
        group         = userProject.group
        give_up       = userProject.give_up
        erralertflag  = userProject.erralertflag
        timeout       = userProject.timeout
        begin_project = userProject.begin_project
        end_project   = userProject.end_project
        file_ready    = userProject.file_ready
        persistent_process = userProject.persistent_process

        # other info:
        status = DEFAULT_FILE_RELEASE_STATUS

	# Establish a consumer
        cid  = sam.establishConsumer(station=sam_station,
                                     project=sam_project,
                                     appName=appname,
                                     appVersion=appversion,
                                     appFamily=appfamily,
                                     group=group)

        if persistent_process == "1":
            fileReader = PersistentFileReceiverAdapter(sam_station,sam_project,cid,erralertflag,timeout)
        else:
            fileReader = TransientFileReceiverAdapter(sam_station,sam_project,cid,erralertflag,timeout)

        # Establish a consumer process
	cpid = fileReader.getCpid()

        # Store these items for easier access
        userProject.setStationName(sam_station)
        userProject.setProjectName(sam_project)
        userProject.setConsumerId(cid)
        userProject.setConsumerProcessId(cpid)

        # set a pointer for ourselves in case we are terminated early.
        self.userProject = userProject

        # execute the user code at project's begin
        result=begin_project()

        nextFileName = fileReader.getNextFile()
        while (self.isFileName(nextFileName)):
            print nextFileName
            ############################################################
            #           USER PROCESSING HAPPENS HERE!                  #
            result = file_ready(nextFileName)                          #
            ############################################################

            if  not self.isError(nextFileName):
              try:
                  sam.releaseFile(station=sam_station, project=sam_project, cpid=cpid,
                                  filename=nextFileName, status=status)
              except SystemExit:
                  # Make sure this exception is propagated if early shutdown is required
                  raise
              except:
                  print "Exception caught releasing the file.  May mean that the file didn't really exist."
            else:
                sys.stderr.write("There was an error getting the file!\n")


            nextFileName = fileReader.getNextFile()

        # execute the user code at project's end

        result=end_project()

        # the 'stop project' command will be handled by the submitter
        print("Project %s has finished." % sam_project)
        return 1

    def isFileName(self, aString):
        if (string.find(aString, "END OF STREAM") != -1): return 0
        return 1

    def isError(self, aString):
        if (string.find(aString, "!ERROR") != -1): return 1
        return 0

################################################################################
#
# if we are running this interactively, then we must assume that
#  SAM_STATION and SAM_PROJECT are set appropriately (because, except
#  for hand-testing, etc., that's the ONLY way we should ever get to
#  the point of running this interactively)

def main(userFile):
    try:
        sam_station = os.environ['SAM_STATION']
        sam_project = os.environ['SAM_PROJECT']
    except KeyError:
        print("ABORT: SAM_STATION and SAM_PROJECT must be set for interactive testing.")

    return RunProject(userFile).run()
    

if __name__ == "__main__":
    #
    # were we called as "python GenericRunProjectScript.py myFile"
    #   or as "GenericRunProjectScript.py myFile"
    #???  It matters.
    # In any case, the onloy argument we pass is the LAST
    # argument, the name of the USER's module.

    userFile = sys.argv[-1]
    
    statusCode = main(userFile)
    if (statusCode):
        sys.exit(0)
    else:
        sys.exit(1)
      
