#!/bin/sh

# This script must not output too much into stdout when it's called
# for job lookup!
prog=`basename $0`
die () { echo "$prog: $@" >&2; exit 1; }
usage () { die Possible actions: `egrep '^[a-z].*\)$' $0 | sed 's?)??g'`; }
get_value () { echo $1 | sed "s?^--[a-z|-]*=??g"; }

checkStatus () { status=`qjob -l | grep $1`; echo $status; }
export PATH="/usr/local/bin:$PATH"

max_job_name_len=20
#unique_suffix=$$
#trying a different way of generating unique suffx as the length of $$ 
#is not constant
unique_suffix=`date | cut -b 12- | cut -b -8 | sed -e 's/://g'`
unique_suffix_len=`echo $unique_suffix | wc -c`
#Precondition: I want to name a job with at least one character
#from the global jid
[ `expr $unique_suffix_len + 1` -gt $max_job_name_len ] && \
  die "Unique job name suffix $unique_suffix is too long! Cannot name a job."

#Arguments:
action=$1
shift
[ -n "$action" ] || usage
#echo $1
while [ -n "$1" ]; do
  case $1 in
  --executable=*)
    #executable=`echo $1 | sed "s?.*=??g"`
    executable=`get_value "$1"`
    echo " The Executable is : "$executable
    ;;
  --arguments=*)
    arguments=`get_value "$1"`
    ;;
  --project=*)
    project=`get_value "$1"`
    project_len=`echo $project | wc -c`
    len=`expr $project_len + $unique_suffix_len`
    
    
    if [ $len -gt $max_job_name_len ]; then
      idx=`expr $len - $max_job_name_len`
      project_name=`echo $project | cut -b ${idx}-`
    else
      project_name=$project
    fi
    #if the 1st character is a digit make it blank otherwise bqs barks
    project_name=`echo $project_name |  sed -e "s/_//" -e 's/^[0-9]/Z/g'`
    ;;
  --stdout=*)
    stdout=`get_value "$1"`
    ;;
  --stderr=*)
    stderr=`get_value "$1"`
    ;;
  --queue=*)
    queue=`get_value "$1"`
    ;;
  --local-job-id=*)
    local_job_id=`get_value "$1"`
    ;;
  *)
    usage
  esac
  shift
done


case $action in

job_submit|project_submit|consumer_submit)
  [ -n "$executable" -a -n "$project" ] || \
      die "specify project and executable when submitting job"

  [ -n "$queue" ] || queue="T"
  
  rfpath="cchpssd0:/hpss/in2p3.fr/group/d0/jim/${project_name}"
  # Check the state of the file on hpss
  cmd="rfstat ${rfpath}"
  eval $cmd
  
  # If the file is not present in the cache on hpss 
  # It means that this is the first submission, set the flag (firstTime)
  
  if [ $? != 0 ];then
     echo "Submitting for the first time"
     firstTime="yes"
     rfcp $executable ${rfpath}
  else
     echo "Subsequent submission"
     firstTime=""
  fi
  
  cmd="echo \"rfcp \${rfpath} $executable; exec sh -c \\\"JIM_LOCAL_JID=\\\$QSUB_REQNAME $executable $arguments\\\" \" | \
         qsub -v SAM_STATION,JIM_JOB_NUMBER -l scratch=4GB,M=1024MB,T=1400000,platform=LINUX,hpss"
  
  #Note: in BQS, unlike PBS, project must be unique
  [ -n "$project_name" ] && cmd="$cmd -N ${project_name}${unique_suffix}"
  [ -n "$stdout" ] && cmd="$cmd -o $stdout"
  [ -n "$stderr" ] && cmd="$cmd -e $stderr"
  [ -n "$queue" ] && cmd="$cmd -q $queue"

  # initialize QSUB_CERTIFICATE
 
  QSUB_CERTIFICATE=`grid-proxy-info -subject` || echo "Could not initialize QSUB_CERTIFICATE"
  export QSUB_CERTIFICATE

  
  echo $cmd
  eval $cmd
  if [ $? != 0 ];then
     echo "$prog:Job Submission failed" >&2
     exit 1
  fi


  #sleep for 1 sec so that the job name is always unique
  sleep 1
  
  # Before declaring that the job is submited , we check 
  # whether the job is visible to the lookup (qjob -l ) command
  # This step makes sure that the job will not fail at the
  # Grid Level even if the BS misbehaves
  if [ -n "$firstTime" ]; then
     echo "$prog: This is the first job submission"
     echo "$prog: Checking the status of the job"
     checkStatus ${project_name}${unique_suffix}
     numTries=1
     while [ -z "$status" -a $numTries -lt 5 ]
     do
        echo "$prog: Looking up for job ${project_name}${unique_suffix}"
	checkStatus ${project_name}${unique_suffix}

        # pick a random number between [0-5]
        random_number=`expr $RANDOM % 6`
        # to get a different number next time echo  to /dev/null
        echo $RANDOM>/dev/null ;
      
        echo "$prog: Random number picked ${random_number}"

        # to get exponential backoff get mul_factor
        mul_factor=`expr $numTries "*" $numTries`

        # 24 - fixed bias to wait atleast 24 secs in 1st case
        sleepTime=`expr 24 + $random_number`
        sleepTime=`expr $sleepTime "*" $mul_factor`
	echo "$prog: Will sleep for  $sleepTime and check again" 
	sleep $sleepTime
	numTries=`expr $numTries + 1`
     done	   
  fi  
      
      
  ;;

job_lookup|consumer_lookup|project_lookup)
   out=`qselect -u \`whoami\` sname  jobname  status`
   if [ $? != 0 ];then
      die "JOB LOOKUP FAILED: ${out}"
   fi
   if [ -z "${out}" ];then
      die "JOB LOOKUP FAILED: output is blank"
   fi

   if [ -n "${local_job_id}" ]; then
      cmd="echo \"\$out\" | fgrep ${local_job_id}"
      #cmd="$cmd ${local_job_id} | tail -1 | fgrep -v 'No job'"
   elif [ -n "$project_name" ]; then 
      #cmd="$cmd | fgrep $project_name"
      cmd="echo \"\$out\" | fgrep $project_name"
   else
      die "Specify local job ID or project name for job lookup"
   fi

   eval $cmd | awk 'BEGIN { statusMap["SIGNALLED"]="suspended" ; \
              	    statusMap["HOLD"]="suspended"; \
                    statusMap["RUNNING"]="active"; \
                    statusMap["STARTED"]="active"; \
                    statusMap["QUEUED"]="pending"; \
                    notFound=1; fmt="JobId=%s Status=%s\n"; } \
                    {  notFound=0; printf fmt, $2, statusMap[$3]; }\
              	    END { if ( notFound ) printf fmt, "UNDEFINED", "done" }'
;;

job_kill|consumer_kill|project_kill)
   out=`qjob -l`
   if [ $? != 0 ];then
      die "JOB KILL FAILED: ${out}"
   fi
   if [ -n "$project_name" ]; then
      local_job_id=`echo "$out" | fgrep $project_name | awk '{print $2}' | tr '\012' ' '`
      echo "Mapped project to job ID's ${local_job_id}"
   elif [ -z "${local_job_id}" ]; then
      die "Specify local job ID or project name for job kill"
   fi
   [ -n "${local_job_id}" ] && qdel ${local_job_id}
   ;;  
  *)
    usage
esac
#HOLD|RUNNING|QUEUED

