#!/usr/bin/env sampy
"""
Internally configure the running environment, then exec the sampy image.
"""

import os
import sys

fileName = sys.path[0]
while os.path.islink(fileName):
    fileName = os.path.normpath(os.path.join(os.path.dirname(fileName),
            os.readlink(fileName)))
dirName = os.path.dirname(fileName)

userlibdirname = os.path.normpath(os.path.join(dirName, '../samlib'))

LD_LIBRARY_PATH = os.environ.get('LD_LIBRARY_PATH','')

if LD_LIBRARY_PATH.startswith(userlibdirname):

    stdpyzipname = os.path.normpath(os.path.join(dirName, '../lib/python%d%d.zip' % sys.version_info[0:2]))
    stdlibpydirname = os.path.normpath(os.path.join(dirName, '../lib/python%d.%d' % sys.version_info[0:2]))
    stdlibdlldirname = os.path.join(stdlibpydirname, 'lib-dynload')
    userpycodefile = os.path.join(userlibdirname,'sam_pyclib.zip')

    sys.frozen = True
    sys.path = [userpycodefile,userlibdirname,fileName,stdpyzipname,stdlibpydirname, stdlibdlldirname]
    
    # set various internal environment variables
    os.environ["SAM_USER_PYAPI_DIR"] = os.path.join(userlibdirname,"sam_user_pyapi")
    os.environ["SAM_ADMIN_PYAPI_DIR"] = os.path.join(userlibdirname,"sam_admin_pyapi")
    os.environ["SAM_MIS_PYAPI_DIR"] = os.path.join(userlibdirname,"sam_mis_pyapi")
    os.environ["SAMGRID_BATCH_ADMIN_PYAPI_DIR"] = os.path.join(userlibdirname,"samgrid_batch_admin_pyapi")
    os.environ["SAMGRID_BATCH_ADAPTER_HANDLER_DIR"] = os.path.join(os.environ["SAMGRID_BATCH_ADMIN_PYAPI_DIR"],"etc/handlers")
    os.environ["SAM_COMMON_PYLIB_DIR"] = "null"
    os.environ["SAM_IDL_PYLIB_DIR"] = "null"

    # frozen sampy has only its own library dir in the path;
    #  add the rest of what the user expects.
    # Ignore directories named pythonx.y that aren't our version 
    PYTHONPATH = os.environ.get('PYTHONPATH')
    if PYTHONPATH:
        import re
        r = re.compile('python\d\.\d')
        thispyver = 'python%d.%d' % sys.version_info[:2]
        for element in PYTHONPATH.split(':'):
            m = r.search(element)
            if not m or  m.group(0) == thispyver:
                sys.path.append(element)

    SAMPYPATH = os.environ.get('SAMPYPATH')
    if SAMPYPATH:
        for element in SAMPYPATH.split(':'):
            sys.path.append(element)
else:
    # set the environment variables and then self exec to
    # pick up the new LD_LIBRARY_PATH
    
    LD_LIBRARY_PATH = '%s:%s' % (userlibdirname, LD_LIBRARY_PATH)
    os.environ['LD_LIBRARY_PATH'] = LD_LIBRARY_PATH
    
    # exec ourself to pick up the changed LD_LIBRARY_PATH
    os.execv(fileName,sys.argv)

# Now import the main module

import zipimport

main = __import__("__main__")
importer = zipimport.zipimporter(fileName)
maincode = importer.get_code(main.__name__)
exec maincode in main.__dict__

# now either run an external script or start an interactive prompt

import code

def _displaytraceback():

    # Stolen from the code module in the std lib
    import traceback
    try:
        type, value, tb = sys.exc_info()
        tblist = traceback.extract_tb(tb)
        del tblist[:1]
        l = traceback.format_list(tblist)
        if l:
            sys.stderr.write("Traceback (most recent call last):\n")
            for e in l:
                sys.stderr.write(e)
        for e in traceback.format_exception_only(type, value):
            sys.stderr.write(e)
    finally:
        tblist = tb = None

def _compile_and_run_code(codefile):
    try:
        exec codefile in main.__dict__
    except KeyboardInterrupt:
        # Don't want a traceback if the user hits ctrl-c
        sys.exit(1)
    except SystemExit:
        # SystemExit should not cause a printed traceback
        raise
    except:
        # display the traceback here to avoid printing the internal frozen sampy frames 
        _displaytraceback()
        sys.exit(1)

# Either execute a python file or start an interpreter, depending on
# command line args
if len(sys.argv) > 1:
    # Modify the args so the script is argv[0], not sampy
    del sys.argv[0]
    scriptfilename = sys.argv[0]

    # put the script directory on the import path
    scriptdir = os.path.abspath(os.path.dirname(scriptfilename))
    sys.path.insert(0,scriptdir)

    try:
        scriptfile = open(scriptfilename,'r')
    except IOError, x:
        print >> sys.stderr, ("%s: can't open file '%s': [Errno %d] %s"
                              % (os.path.basename(sys.executable), x.filename, x.errno, x.strerror))
        sys.exit(2)
    _compile_and_run_code(scriptfile)

else:
    # no script on the command line.
    sys.argv[0] = ''

    # put the current directory on the import path
    sys.path.insert(0,'')
    # if std in is a terminal then start an interactive prompt
    if os.isatty(sys.stdin.isatty()):
        class _Helper(object):
            """Define the built-in 'help'.
            This is a wrapper around pydoc.help (with a twist).
            
            """
            def __repr__(self):
                return "Type help() for interactive help, " \
                       "or help(object) for help about object."
            def __call__(self, *args, **kwds):
                import pydoc
                return pydoc.help(*args, **kwds)
        main.help = _Helper()
        from sampy_version import sampy_version
        code.interact("Welcome to SAM\nsampy %s using Python %s on %s" % (sampy_version, sys.version, sys.platform), local=main.__dict__)
    else:
        # assume we're getting code from std input
        _compile_and_run_code(sys.stdin)
        

        
