#!/usr/bin/env python

# Standard python modules.
import sys
import re
import os
import types
import time

# SAM IDL stuff.
import SAM

# Other SAM stuff.
import DbServerProxy
import SamCpFileParser
import ModuleLoader
import CommonUtility
import ExceptionManager
import SAMError

# SAM CP stuff.
from SamCpProtocolArbitrator import SamCpProtocolArbitrator
import SamCpExceptions

##############################################################################
# Constants.

TRUE_ = 1
FALSE_ = 0

DEFAULT_TIMESTAMP_FORMAT_ = "<%m/%d/%y %H:%M:%S>"

VERBOSE_FLAG_ = "v"
DEBUG_FLAG_ = "d"

PROTOCOL_CONSTRAINT_MAP_ = "ProtocolConstraintMap_"
DEFAULT_PROTOCOL_ = "DefaultProtocol_"

CONFIG_MODULE_ENV_VAR_ = "SAM_CP_ARBITRATOR_CONFIG_MOD"
CONFIG_DIR_ENV_VAR_ = "SAM_CP_ARBITRATOR_CONFIG_DIR"

##############################################################################
# Exception classes.

# Generic error class.
class BaseModuleException(SamCpExceptions.SamCpException):
  
  def __init__(self, error = None):
    """ Initialization. """
    SamCpExceptions.SamCpException.__init__(self, error)

# Invalid configuration module.
class InvalidConfigurationModule(BaseModuleException):
  
  def __init__(self, error = None):
    """ Initialization. """
    BaseModuleException.__init__(self, error)

# Initialization error.
class InitializationError(BaseModuleException):
  
  def __init__(self, error = None):
    """ Initialization. """
    BaseModuleException.__init__(self, error)

# Invalid protocol constraint.
class InvalidProtocolConstraint(BaseModuleException):
  
  def __init__(self, error = None):
    """ Initialization. """
    BaseModuleException.__init__(self, error)


#############################################################################
# Utility functions.

# Logging.
def log(what, verboseFlag = TRUE_):
  """ Log string into a log file. """
  if(verboseFlag == TRUE_):
    # Prepare timestamp.
    t = time.time()
    timeStamp = time.strftime(DEFAULT_TIMESTAMP_FORMAT_, time.localtime(t))

    # Prepare string and print.
    outputStr = "%s %s\n" % (timeStamp, what)
    sys.stdout.write(outputStr)
    sys.stdout.flush()

##############################################################################
# File MetaData Handler class.

class FileMetaDataHandler:

  class BaseClassException(BaseModuleException):

    def __init__(self, error = None):
      """ Initialization. """
      if(error):
        self.args = "%s" % error

  def __init__(self, fileName):
    self._fileName = fileName
    self._metaData = None
    try:
      dbs = DbServerProxy.DbServerProxy("File").proxy()
      self._metaData = dbs.getFileMetaDataWithCRC(self._fileName)
    except SAM.Exception, ex:
      errMsg = "SAM Exception caught: " + ex.content.v.additionalInformation \
               + " \n%s" % (
                 CommonUtility.PrependOutput(
                   "  ", ExceptionManager.formatExceptionInfo()))
      raise FileMetaDataHandler.BaseClassException(errMsg)
    except:
      raise
    self._metaDataDict = self.flattenDict()

  def getFileName(self):
    return self._fileName

  def getMetaData(self):
    return self._metaData

  def getMetaDataDict(self):
    return self._metaDataDict;
  
  def flattenDict(self, object = None):
    if(object == None):
      return self.flattenDict(self._metaData.v)
    else:
      result = {}
      dictKeys = []
      dictKeys = object.__dict__.keys()
      for k in dictKeys:
        o = object.__dict__[k]
        try:
          oDict = self.flattenDict(o)
          for k2 in oDict.keys():
            result[k2] = oDict[k2]
        except AttributeError:
          result[k] = o
      return result

##############################################################################
# File transfer protocol handler class.

class SamCpMetadataProtocolArbitrator(SamCpProtocolArbitrator):

  def __init__(self, *args):
    SamCpProtocolArbitrator.__init__(self, *args)
    srcFile = self.getSrcLoc()
    destFile = self.getDestLoc()
    
    try:
      self._protocolConstraintAnalyzer = ProtocolConstraintAnalyzer(srcFile,
                                                                    destFile)
      self._configManager = getConfigManagerInstance()
    except BaseModuleException, ex:
      errMsg = "Cannot initialize SamCpMetadataProtocolArbitrator: \n%s" % (
        CommonUtility.PrependOutput(
          "  ", ExceptionManager.formatExceptionInfo()))
      raise InitializationError(errMsg)


  def getProtocolConstraintAnalyzer(self):
    return self._protocolConstraintAnalyzer

  def selectProtocol(self):
    protocolConstraintMap = self._configManager.getProtocolConstraintMap()
    protocol = self._configManager.getDefaultProtocol()
    protocolMsg = "Using default protocol: %s" % (protocol)
    match = FALSE_
    for p in protocolConstraintMap.keys():
      constraintList = protocolConstraintMap[p]
      if(type(constraintList) != types.ListType):
        msg = "Ignoring invalid constraint list: %s (protocol: %s)" % (
          constraintList, p)
        self.dbgprint(msg)
      else:
        if(len(constraintList)):
          for c in constraintList:
            if(type(c) != types.DictType):
              msg = "Ignoring invalid constraint: %s (protocol: %s)" % (c, p)
              self.dbgprint(msg)
            else:
              msg = "Analyzing constraint %s (protocol: %s)" % (c, p)
              self.dbgprint(msg)
              try:
                match = self._protocolConstraintAnalyzer.match(c)
                break
              except ProtocolConstraintAnalyzer.ConstraintMatchFailure, ex:
                self.dbgprint("%s" % (ex))
      if(match):
        protocol = p 
        protocolMsg = "Matched protocol: %s" % (p)
        break

    self.dbgprint(protocolMsg)
    return protocol

##############################################################################
# Constraint analyzer class.

class ProtocolConstraintAnalyzer:

  ############################################################################
  # Class exceptions.
  class BaseClassException(BaseModuleException):

    def __init__(self, error = None):
      """ Initialization. """
      if(error):
        self.args = "%s" % error

  class ConstraintMatchFailure(BaseClassException):

    def __init__(self, error = None):
      """ Initialization. """
      if(error):
        self.args = "%s" % error
    
  def __init__(self, srcFile, destFile):
    self._srcFile = srcFile
    self._destFile = destFile
    self._fileName = srcFile.basename()
    self._fileMetaDataHandler = FileMetaDataHandler(self._fileName)
    self._metaDataDict = self._fileMetaDataHandler.getMetaDataDict()

  def getFileName(self):
    return self._fileName

  def getFileMetaDataHandler(self):
    return self._fileMetaDataHandler

  def match(self, constraintDict):
    if(type(constraintDict) != types.DictType):
      raise SAMError.ArgumentTypeError("dict", "constraintDict", 
                                       constraintDict)
    result = TRUE_
    for key in constraintDict.keys():
      keyStr = "%s" % (key)
      valueStr = "%s" % (constraintDict[key])
      if(self._metaDataDict.has_key(keyStr)):
        metaDataValueStr = self._metaDataDict[keyStr]
        if(not re.search(valueStr, metaDataValueStr)):
          msg = "File meta data value \"%s\" for key \"%s\" " % (
            metaDataValueStr, keyStr) \
            + "does not match the constraint value \"%s\". " % (valueStr)
          raise ProtocolConstraintAnalyzer.ConstraintMatchFailure(msg)
      else:
        msg = "No key \"%s\" in the file meta data." % (keyStr)
        raise ProtocolConstraintAnalyzer.ConstraintMatchFailure(msg)
    return result
      
      
##############################################################################
# Config manager singleton class.

class ConfigManager:

  __instance = None

  # Initialization.
  def __init__(self, configModuleName, configModuleDir):
    """ Initialization. """
    
    # Singleton stuff.
    if( ConfigManager.__instance):
      raise ConfigManager.__instance
    ConfigManager.__instance = self

    # Import configuration module.
    try:
      self._configModuleName = configModuleName
      self._configModuleDir = configModuleDir
      self._configModule = ModuleLoader.Load(configModuleName, configModuleDir)
      cmd = "self._protocolConstraintMap = self._configModule.%s" % (
        PROTOCOL_CONSTRAINT_MAP_)
      exec cmd
      if(type(self._protocolConstraintMap) != types.DictType):
        msg = "%s should be a dictionary "  % (PROTOCOL_CONSTRAINT_MAP_) \
              + "of the form protocol : constraintList"

      cmd = "self._defaultProtocol = self._configModule.%s" % (
        DEFAULT_PROTOCOL_)
      exec cmd

    except InvalidConfigurationModule, ex:
      raise
    except:
      err = "Invalid configuration module: \n%s" % (
        CommonUtility.PrependOutput(
          "  ", ExceptionManager.formatExceptionInfo()))
      raise InvalidConfigurationModule(err)

  # Return protocol constraint map.
  def getProtocolConstraintMap(self):
    """ Return protocol constraint map. """
    return self._protocolConstraintMap

  # Return default protocol.
  def getDefaultProtocol(self):
    """ Return default protocol. """
    return self._defaultProtocol

##############################################################################
# Config manager singleton instance.

def getConfigManagerInstance(configModuleName = None, 
                             configModuleDir = None):
  """ Return  config manager. """ 
  if(configModuleName is None):
    try:
      configModuleName = os.environ[CONFIG_MODULE_ENV_VAR_]
    except KeyError, ex:
      error = "No configuration module provided."
      raise InvalidConfigurationModule(error)
  
  if(configModuleDir is None):
    try:
      configModuleDir = os.environ[CONFIG_DIR_ENV_VAR_]
    except KeyError, ex:
      # will try without it.
      pass
  
  cm = None
  try:
    cm = ConfigManager(configModuleName, configModuleDir)
  except ConfigManager, cman:
    cm = cman

  return cm

##############################################################################
# Unit tests.

if(__name__ == "__main__"):

  srcFile = "datalogger://jerry/sam_demo_file.ev"
  destFile = "enstore://this_is_my_dir/sam_demo_file.ev"
  configDir = "/scratch/d0test/veseli/sam_devel/online"
  configModule = "sam_cp_online_config"

  print "Initializing FileTransferProtocolHandler"
  protocolHandler = FileTransferProtocolHandler(srcFile, destFile) 
  protocolHandler.determineProtocol()

  #configManager = getConfigManagerInstance(configModule, configDir)
  #print configManager.getProtocolConstraintMap()

  #stream = metadata.v.commonFileMetaData.physicalStream
  #$print metadata.v.__dict__.keys()
  #print "Stream = ", stream
  #match = re.search("electron.*", stream)
  #print match.start()
  #print match.end()
  #print match.group()
  print "Done testing"    

