#!/usr/bin/env sampy
import os
import sys
import string

from SamCpImportManager import getSamCpImportManager
from SamCpDomainHandler import SamCpDomainHandler

from SamCpExceptions import SamCpMethodNotImplementedException, SamCpUnrecognizedProtocol
        
class SamCpConfigManager:
    """
    Class to handle sam_cp configuration.
    """
    #################################################################
    def __init__(self, samCpCommand):
        self.configFile      = samCpCommand.getConfigFile()
        self.useThisProtocol = samCpCommand.getUseThisProtocol()
        self.curLoc          = samCpCommand.getCurLoc()
        self.srcLoc          = samCpCommand.getSrcLoc()
        self.destLoc         = samCpCommand.getDestLoc()
        self.debug           = samCpCommand.getDebug()
        self.dbgprint        = samCpCommand.getDbgprint()
   #################################################################
    def getKnownSamCpProtocolNames(self):
        """
        returns a list of strings corresponding to known protocol names
        """
        #############################################################
        # MUST BE IMPLEMENTED IN THE DERIVED CLASS!
        raise SamCpMethodNotImplementedException()
    #################################################################
    def getLocallyAvailableSamCpProtocolNames(self):
        """
        returns a list of strings corresponding to protocol names that
        are available locally
        """
        #############################################################
        # MUST BE IMPLEMENTED IN THE DERIVED CLASS!
        raise SamCpMethodNotImplementedException()
    #################################################################
    def getPreferredSamCpProtocolName(self):
        """
        returns the locally preferred protocol name
        """
        #############################################################
        # MUST BE IMPLEMENTED IN THE DERIVED CLASS!
        raise SamCpMethodNotImplementedException()
    #################################################################
    def getProtocolClass(self, protocolName):
        """
        returns the class associated with a known protocol
        """
        #############################################################
        # MUST BE IMPLEMENTED IN THE DERIVED CLASS!
        raise SamCpMethodNotImplementedException()
###############################################################################

class SamCpDefaultConfigManager(SamCpConfigManager):
    KNOWN_SAM_CP_CLASSES_ = "KNOWN_SAM_CP_CLASSES"
    DOMAIN_CAPABILITY_MAP_ = "DOMAIN_CAPABILITY_MAP"
    DEFAULT_SAM_CP_PROTOCOL_NAME_ = "rcp"
    
    def __init__(self, samCpCommand):
        SamCpConfigManager.__init__(self, samCpCommand)

        self.knownSamCpProtocolNames = None
        self.locallyAvailableSamCpProtocolNames = None
        self.preferredSamCpProtocolName = None

        self.protocolMap = None
        self.domainCapabilityMap = None

        self.initialize()
        
    #################################################################
    def initialize(self):
        # default config manager uses the $SAM_CP_CONFIG_FILE to
        # determine the list of known protocols, domains, etc.

        importManager = getSamCpImportManager()
        self.protocolMap = importManager.importObjectFromFile(self.KNOWN_SAM_CP_CLASSES_, self.configFile)
        self.domainCapabilityMap = importManager.importObjectFromFile(self.DOMAIN_CAPABILITY_MAP_, self.configFile)

        # the knownSamCpProtocols are all of those listed in the protocolMap
        self.knownSamCpProtocolNames = self.protocolMap.keys()
                                                              
        # determine which protocols are available to the local node
        # based on the domainCapabilityMap
        self.locallyAvailableSamCpProtocolNames = []

        if ( self.useThisProtocol ):
            # if the user specifically requested a particular protocol,
            # then the name of the protocol should be the ONLY locally available
            # protocol that we list!
            self.locallyAvailableSamCpProtocolNames.append(self.useThisProtocol)
        else:
            # start looking for which transport mechanisms are available.
            # easy cases first.
            # We'll build a list, because maybe some day in the
            # future we'll want to try more than one.  The order
            # of the list will be the order in which to try the
            # different protocols.
            #
            # if everything is in the same domain, we use local copy (cp)
            if ( self.curLoc.domain() == self.srcLoc.domain() and
                 self.curLoc.domain() == self.destLoc.domain() ):
                self.locallyAvailableSamCpProtocolNames.append('local_cp')

            # now look in the $SAM_CP_CONFIG_FILE map

            # the parsed domains may not be the same as the strings
            # in the configuration file (e.g., the currentDomain may be
            # "d0ora3.fnal.gov", while the map file may have configured
            # only "fnal.gov".

            # create an instance of the domain parser/handler
            domainHandler = SamCpDomainHandler(self.domainCapabilityMap, self.dbgprint)

            # Build a list of the "best matches":
            configuredDomains = ( domainHandler.samCpDomain(self.curLoc.domain()),
                                  domainHandler.samCpDomain(self.srcLoc.domain()),
                                  domainHandler.samCpDomain(self.destLoc.domain()) )

            # and from the list of domains that are configured, build
            #  the list of available protocol keywords
            domainHandlerList = domainHandler.getAvailableProtocols(configuredDomains)
            for protocolName in domainHandlerList:
                if ( protocolName not in self.locallyAvailableSamCpProtocolNames and
                     protocolName in self.knownSamCpProtocolNames ):
                    self.locallyAvailableSamCpProtocolNames.append(protocolName)

        # if we haven't found a specific available protocol,
        # fall back to the default protocol keyword:
        if ( len(self.locallyAvailableSamCpProtocolNames) == 0 ):
            self.dbgprint("Warning, no defined protocols match request; defaulting to '%s'." % self.DEFAULT_SAM_CP_PROTOCOL_NAME_)
            self.locallyAvailableSamCpProtocolNames.append(self.DEFAULT_SAM_CP_PROTOCOL_NAME_)

        # the preferred protocol is, by default, the first one in the list.
        self.preferredSamCpProtocolName = self.locallyAvailableSamCpProtocolNames[0]
            
        self.dbgprint("knownSamCpProtocolNames = %s" % str(self.knownSamCpProtocolNames))
        self.dbgprint("locallyAvailableSamCpProtocolNames = %s" % str(self.locallyAvailableSamCpProtocolNames))
        self.dbgprint("preferredSamCpProtocolName = %s" % str(self.preferredSamCpProtocolName))
        self.dbgprint("all known protocols: %s" % str(self.protocolMap))
        
    #################################################################
    def getKnownSamCpProtocolNames(self):
        """
        returns a list of strings corresponding to known protocol names
        """
        return self.knownSamCpProtocolNames
    #################################################################
    def getLocallyAvailableSamCpProtocolNames(self):
        """
        returns a list of strings corresponding to protocol names that
        are available locally
        """
        return self.locallyAvailableSamCpProtocolNames
    #################################################################
    def getPreferredSamCpProtocolName(self):
        """
        returns the locally preferred protocol name
        """
        return self.preferredSamCpProtocolName
    #################################################################
    def getProtocolClass(self, protocolName):
        """
        returns the class associated with a known protocol
        """
        try:
            return self.protocolMap[protocolName]
        except KeyError:
            raise SamCpUnrecognizedProtocol(protocolName)
