#!/usr/bin/env sampy
###############################################################################
# Project : SAM
# Package : sam_cp
# $Id: SamCpEnstoreFileInfoHandler.py,v 1.7 2005/06/14 17:14:38 lauri Exp $
#
# This work is part of a development project, called SAM, which consists of a
# number of coordinated packages each named sam_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
###############################################################################

import types
import sys
import string

import SamCpExceptions

from SamUtility.CommonUtility import ExecSubprocess
from SamException.SamExceptionFormatTools import formatExceptionInfo

###############################################################################
# Constants.

ENSTORE_FILE_INFO_CMD_ = "fileinfo"
ENSTORE_FILE_INFO_KEY_ = "file_info"
ENSTORE_VOLUME_INFO_KEY_ = "volume_info"
ENSTORE_CRC_VALUE_KEY_ = "complete_crc"
DCACHE_CRC_VALUE_KEY_ = "l2crc"
ENSTORE_VOLUME_LABEL_KEY_ = "external_label"
ENSTORE_VOLUME_OFFSET_KEY_ = "location_cookie"

ENSTORE_VOLUME_OFFSET_DELIMITER_ = "_"

UNKNOWN_CRC_VALUE_ = "unknown crc value"
UNKNOWN_VOLUME_OFFSET_ = "unknown volume offset"
UNKNOWN_VOLUME_LABEL_ = "unknown volume label"


###############################################################################
# Exception classes.

class SamCpEnstoreFileInfoHandlerException(SamCpExceptions.SamCpException):

  def __init__(self, error = ""):
    self.args = error

class UnknownInformationRequested(SamCpEnstoreFileInfoHandlerException):

  def __init__(self, *args):
    SamCpEnstoreFileInfoHandlerException.__init__(self, *args)

class UnknownFile(SamCpEnstoreFileInfoHandlerException):

  def __init__(self, *args):
    SamCpEnstoreFileInfoHandlerException.__init__(self, *args)

class UnknownFileInfo(SamCpEnstoreFileInfoHandlerException):

  def __init__(self, *args):
    SamCpEnstoreFileInfoHandlerException.__init__(self, *args)

class UnknownVolumeInfo(SamCpEnstoreFileInfoHandlerException):

  def __init__(self, *args):
    SamCpEnstoreFileInfoHandlerException.__init__(self, *args)

###############################################################################
# File info handler class.

class SamCpEnstoreFileInfoHandler:


  def __init__(self, fileName):
    if type(fileName) != types.StringType:
      err = "%s: Invalid argument type." % (self.__class__)
      raise SamCpEnstoreFileInfoHandlerException(err)
    self._fileName = fileName
    self._fileInfoDict = None
    self._volumeInfoDict = None
    
  def updateInfo(self):
    try:
      cmd = "%s %s" % (ENSTORE_FILE_INFO_CMD_, self._fileName)
      (stdOut, stdErr, exitStatus) = ExecSubprocess(cmd)
      
      if exitStatus:
        err = "%s: Unable to update file info. \n" % (self.__class__) \
              + "StdOut: \n%s\n" % (stdOut) \
              + "StdErr: \n%s\n" % (stdErr) \
              + "ExitStatus: \n%s\n" % (exitStatus) 
        raise SamCpEnstoreFileInfoHandlerException(err)
      
      cmd = "dict = %s" % (stdOut)
      exec cmd
      fileDict = dict[self._fileName] 
    except:
      err = formatExceptionInfo()
      msg = "%s: Could not update file info: %s" % (self.__class__, err)
      raise SamCpEnstoreFileInfoHandlerException(err)

    if not fileDict:
      err = "%s: No information available for file %s." % (self.__class__,
                                                           self._fileName)
      raise UnknownFileInfo(err)
       

    # Try first the L2 CRC only.
    if fileDict.has_key(DCACHE_CRC_VALUE_KEY_):
      self._fileInfoDict = {}
      self._fileInfoDict[ENSTORE_CRC_VALUE_KEY_] = fileDict[DCACHE_CRC_VALUE_KEY_]
    # Try reading the complete file info.
    if fileDict.has_key(ENSTORE_FILE_INFO_KEY_):
      if self._fileInfoDict is None:
        self._fileInfoDict = fileDict[ENSTORE_FILE_INFO_KEY_]
      else:
        for k in fileDict.keys():
          self._fileInfoDict[k] = fileDict[k]

    # If nothing worked, throw exception.
    if self._fileInfoDict is None:
      err = "%s: Unknown file info for file %s." % (self.__class__,
                                                    self._fileName)
      raise UnknownFileInfo(err)

    # Try reading the complete volume info.
    if fileDict.has_key(ENSTORE_VOLUME_INFO_KEY_):
      self._volumeInfoDict = fileDict[ENSTORE_VOLUME_INFO_KEY_]
    else:
      err = "%s: Unknown volume info for file %s." % (self.__class__,
                                                      self._fileName)
      raise UnknownVolumeInfo(err)

  def getFileInfoDict(self):
    if self._fileInfoDict is None:
      try:
        self.updateInfo()
      except UnknownVolumeInfo, ex:
        # We do not need volume info here.
        pass
    return self._fileInfoDict

  def getVolumeInfoDict(self):
    if self._volumeInfoDict is None:
      self.updateInfo()
    return self._volumeInfoDict

  def getCRCValue(self):
    fiDict = self.getFileInfoDict()
    result = UNKNOWN_CRC_VALUE_
    # Add "L" for backwards compatibility.
    # This should be removed once crc checks go into sam_cp.
    if fiDict.has_key(ENSTORE_CRC_VALUE_KEY_):
      result = "%sL" % (long(fiDict[ENSTORE_CRC_VALUE_KEY_]))
    return result

  def getVolumeLabel(self):
    fiDict = self.getFileInfoDict()
    result = UNKNOWN_VOLUME_LABEL_
    if fiDict.has_key(ENSTORE_VOLUME_LABEL_KEY_):
      result = fiDict[ENSTORE_VOLUME_LABEL_KEY_]
    return result

  def getVolumeOffset(self):
    fiDict = self.getFileInfoDict()
    result = UNKNOWN_VOLUME_OFFSET_
    if fiDict.has_key(ENSTORE_VOLUME_OFFSET_KEY_):
      cookie = fiDict[ENSTORE_VOLUME_OFFSET_KEY_]
      cookie = string.split(cookie, ENSTORE_VOLUME_OFFSET_DELIMITER_)[-1]
      result = "%s" % (long(cookie))
    return result

  def getValue(self, key):
    keyStr = "%s" % key
    fiDict = self.getFileInfoDict()
    if fiDict.has_key(keyStr):
      return fiDict[keyStr]
    else:
      err = "%s: Unknown information requested: %s" % (self.__class__, keyStr)
      raise UnknownInformationRequested(err)
    
###############################################################################
# Unit tests.

if __name__ == "__main__":

  fh = SamCpEnstoreFileInfoHandler("/pnfs/sam/dzero/demo/sam_demo_file.ev")
  fh.getFileInfoDict()
  print fh.getCRCValue()
  print fh.getVolumeLabel()
  print fh.getVolumeOffset()
  print fh.getValue("status")
  print fh.getValue("xxx")
