#
# sam_cp_config.py
#
# python module which gives the mappings for
# different ways to transfer files.
#

import os,sys
import copy
import SamCpClasses

global KNOWN_SAM_CP_CLASSES
global DOMAIN_CAPABILITY_MAP


#
# Known methods for handling the different types of Sam Copy protocols.
# The dictionary key is a string which determines the class to use
# for performing the copy.  The classes are defined in
# SamCpClasses.py.
#

KNOWN_SAM_CP_CLASSES = { 'sam_kerberos_rcp' : SamCpClasses.SamKerberosRcp,
                         'local_cp'         : SamCpClasses.SamLocalCp,
                         'rcp'              : SamCpClasses.SamRcp,
                         'scp'              : SamCpClasses.SamScp,
                         'enstore'          : SamCpClasses.SamEnstore,
                         'encp'             : SamCpClasses.SamEncp,
                         'dcache'           : SamCpClasses.SamDcache,
                         'dcache_gridftp'   : SamCpClasses.SamDcacheGridftp,
                         'sam_gridftp'      : SamCpClasses.SamGridftp,
                         'jim_gridftp'      : SamCpClasses.JimGridftp,
                         'dccp'             : SamCpClasses.SamDccp,
                         'GridKaCp'         : SamCpClasses.SamGridKaCp,
                         'rfio'             : SamCpClasses.SamRfio,
                         'bbrfio'           : SamCpClasses.SamBbftpRfio,
                         'sam_bbftp'        : SamCpClasses.SamBbftp,
                         }



#
# Mapping between a DOMAIN name and the SAM CP CLASSES
# that it provides.
#
# If multiple matches are found, the priority will
# be given to EARLIER entries in the INITIATING
# domain's entry.
#
DOMAIN_CAPABILITY_MAP = { 'enstore'      : [ 'enstore',  \
                                             'encp', \
                                             'dcache', \
                                             'dccp', \
                                             'dcache_gridftp', ],
                          'fnal.gov'     : [ 'sam_kerberos_rcp', \
                                             'rcp', \
                                             'enstore', 'encp', 'dcache', 'bbrfio', ],
                          'man.ac.uk'    : [ 'sam_gridftp', ],
                          'cs.wisc.edu'  : [ 'sam_gridftp', ],
                          'dummyDomain'  : [ 'dummyClass', ],
                          'in2p3.fr'     : [ 'rcp', \
                                             'rfio', \
                                             'sam_bbftp', 'bbrfio', ] ,
                          'rfio://in2p3.fr': [ 'rfio', \
                                               'bbrfio', ],
                          'YOUR_NODE_HERE' : [],
                          }


########################################################################
#
# Template for adding new SamCp classes:
#  a) create the Class with methods __init__(self) and
#     copy(src,dest) which handles this new protocol
#  b) add a string description dictionary key and the name
#     of the path to the KNOWN_SAM_CP_CLASSES dictionary


# New class definition,
#  overriding the default __init__() and copy() methods.
#
#
class DummySamCpClass(SamCpClasses.SamCp):
    def __init__(self):
        SamCp.__init__(self)

    # The guts of the matter take place here.
    #
    # srcFileObject and destFileObjects
    # are described in the file
    # $SAM_CP_DIR/src/SamCpFileParser.py
    #
    # The known classes are defined in the
    # file
    # $SAM_CP_DIR/src/SamCpClasses.py.
    #
    def copy(self, srcFileObject, destFileObject):
        SamCp.copy(self, srcFileObject, destFileObject)


# Add the new class definition to the
# KNOWN_SAM_CP_CLASSES dictionary, so that
# any sites that have 'dummyClass' in their
# DOMAIN_PROTOCOL_MAP entry know which class
# to call to perform this transfer.
KNOWN_SAM_CP_CLASSES['dummyClass'] = DummySamCpClass
        



